
local MAJOR = "StarLibProperty-1.0" 
local MINOR = 13
assert(LibStub, MAJOR.." requires LibStub") 
local LibProperty = LibStub:NewLibrary(MAJOR, MINOR)
local libEval = "StarLibEvaluator-1.0"
assert(LibStub(libEval), MAJOR .. " requires " .. libEval)
local Evaluator = LibStub(libEval):New()
local errorHandler
assert(LibStub("StarLibError-1.0"), MAJOR .. " requires StarLibError-1.0")

local pool = setmetatable({},{__mode='k'})

if not LibProperty or not Evaluator then return end

if not LibProperty.pool then
	LibProperty.__index = LibProperty
end

-- Create a new StarLibProperty object
-- @name StarLibProperty.New
-- @usage LibProperty:New(env, name, expression, defval)
-- @param visitor A StarLibCore object
-- @param env The execution environment
-- @param name Give your property a name. 
-- @param expression The lua script for this property.
-- @param defval The default value for this property.
-- @return A new StarLibProperty object
function LibProperty:New(widget, visitor, name, expression, defval, errorLevel)
	assert(visitor.environment, "Unable to create new property. Requires an environment")
	assert(name, "Unable to create new property. Requires a name")
	
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
	end

	setmetatable(obj, self)

	obj.visitor = visitor
	obj.name = name
	obj.environment = visitor.environment
	obj.widget = widget
	obj.expression = expression
	obj.defval = defval
	obj.errorLevel = errorLevel

	obj.is_valid = true
	obj.error = LibStub("StarLibError-1.0"):New(MAJOR, errorLevel)
	
	if type(expression) ~= "string" then 
		obj.is_valid = false; 
	else
		obj.result = Evaluator.ExecuteCode(obj.environment, name, expression, false, defval)
		if obj.result == nil then
			obj.error:Print(("Property invalid: expression = \"%s\""):format(expression))
			obj.is_valid = false
		end
	end
	return obj	
end

-- Delete a property
-- @name StarLibPropoerty.Del
-- @usage LibProperty:Del([prop])
-- @param prop An optional property.
-- @return Nothing
function LibProperty:Del(prop)
	if not prop then prop = self end
	pool[prop] = true
	prop.error:Del()
	prop.error = nil
end

-- Evaluate a property's code
-- @name StarLibProperty.Eval
-- @usage object:Eval()
-- @return Nothing
function LibProperty:Eval()
	if not self.is_valid then return 0 end
	
	local update = 1
	
	local old = self.result
	
	self.environment.self = self.widget
	self.result = Evaluator.ExecuteCode(self.environment, self.name, self.expression, false, self.defval)
	self.environment.self = nil
	
	if old == self.result then
		update = 0
	end
	
	return update
end

-- Return the property's value as a number
-- @name StarLibProperty.P2N
-- @usage object:P2N()
-- @return Nothing
function LibProperty:P2N()
	if not self.is_valid then return self.defval end
	if type(self.result) == "number" then
		return self.result
	else
		return tonumber(self.result)
	end
end

-- Return the property's value as a string
-- @name StarLibProperty.P2S
-- @usage object:P2N()
-- @return Nothing
function LibProperty:P2S()
	if not self.is_valid then return self.defval end
	
	if type(self.result) ~= "number" and type(self.result) ~= "string" then
		return ""
	end
	if type(self.result) == "number" then
		return ("%d"):format(self.result)
	else
		return self.result
	end
end
